-----------------------------------------------------------------------------------------
{-| Module      :  Database.HSQL.SQLite3
    Copyright   :  (c) Krasimir Angelov 2005
    License     :  BSD-style

    Maintainer  :  kr.angelov@gmail.com
    Stability   :  provisional
    Portability :  portable

    The module provides interface to SQLite3
-}
-----------------------------------------------------------------------------------------

module Database.HSQL.SQLite3(connect, module Database.HSQL) where

import Database.HSQL
import Database.HSQL.Types
import Foreign
import Foreign.C
import System.IO
import Control.Monad(when)
import Control.Exception(throwDyn)
import Control.Concurrent.MVar

#include <fcntl.h>
#include <sqlite3.h>

type SQLite3 = Ptr ()
type SQLite3Stmt = Ptr ()

foreign import ccall sqlite3_open :: CString -> (Ptr SQLite3) -> IO Int
foreign import ccall sqlite3_errmsg :: SQLite3 -> IO CString
foreign import ccall sqlite3_close :: SQLite3 -> IO ()
foreign import ccall sqlite3_exec :: SQLite3 -> CString -> FunPtr () -> Ptr () -> Ptr CString -> IO CInt
foreign import ccall sqlite3_prepare :: SQLite3 -> CString -> CInt -> Ptr SQLite3Stmt -> Ptr CString -> IO CInt
foreign import ccall sqlite3_finalize :: SQLite3Stmt -> IO CInt
foreign import ccall sqlite3_step :: SQLite3Stmt -> IO CInt
foreign import ccall sqlite3_reset :: SQLite3Stmt -> IO CInt
foreign import ccall sqlite3_changes :: SQLite3 -> IO CInt
foreign import ccall sqlite3_column_count :: SQLite3Stmt -> IO CInt
foreign import ccall sqlite3_column_name :: SQLite3Stmt -> CInt -> IO CString
foreign import ccall sqlite3_column_text :: SQLite3Stmt -> CInt -> IO CString
foreign import ccall sqlite3_column_type :: SQLite3Stmt -> CInt -> IO CInt
foreign import ccall sqlite3_bind_double :: SQLite3Stmt -> CInt -> Double -> IO CInt
foreign import ccall sqlite3_bind_int :: SQLite3Stmt -> CInt -> CInt -> IO CInt
foreign import ccall sqlite3_bind_int64 :: SQLite3Stmt -> CInt -> Int64 -> IO CInt
foreign import ccall sqlite3_bind_null :: SQLite3Stmt -> CInt -> IO CInt
foreign import ccall sqlite3_bind_text :: SQLite3Stmt -> CInt -> CString -> CInt -> CInt -> IO CInt
foreign import ccall sqlite3_get_table ::   SQLite3 -> CString -> Ptr (Ptr CString) -> Ptr CInt -> Ptr CInt -> Ptr CString -> IO CInt
foreign import ccall sqlite3_free_table :: Ptr CString -> IO ()
foreign import ccall sqlite3_free :: CString -> IO ()

foreign import ccall "strlen" strlen :: CString -> IO CInt

-----------------------------------------------------------------------------------------
-- routines for handling exceptions
-----------------------------------------------------------------------------------------

handleSqlResult :: SQLite3 -> CInt -> IO ()
handleSqlResult sqlite res
	| res == (#const SQLITE_OK) = return ()
	| otherwise = do
            pMsg <- sqlite3_errmsg sqlite
            msg <- peekCString pMsg
            sqlite3_free pMsg
            throwDyn (SqlError "E" (fromIntegral res) msg)

-----------------------------------------------------------------------------------------
-- Connect
-----------------------------------------------------------------------------------------

connect :: FilePath -> IOMode -> IO Connection
connect fpath mode =
	alloca $ \psqlite ->
	  withCString fpath $ \pFPath -> do
		res <- sqlite3_open pFPath psqlite
		sqlite <- peek psqlite
		when (res /= (#const SQLITE_OK)) $ do
			pMsg <- sqlite3_errmsg sqlite
			msg <- peekCString pMsg
			throwDyn (SqlError
			            { seState = "C"
			            , seNativeError = res
			            , seErrorMsg = msg
			            })
		refFalse <- newMVar False
		let connection = Connection
			{ connDisconnect = sqlite3_close sqlite
			, connClosed     = refFalse
			, connExecute    = execute sqlite
			, connPrepare    = prepare connection sqlite
			, connQuery      = query connection sqlite
			, connTables     = tables connection sqlite
			, connDescribe   = describe connection sqlite
			, connBeginTransaction = execute sqlite "BEGIN TRANSACTION" >> return ()
			, connCommitTransaction = execute sqlite "COMMIT TRANSACTION" >> return ()
			, connRollbackTransaction = execute sqlite "ROLLBACK TRANSACTION" >> return ()
			}
		return connection
	where
		oflags1 = case mode of
	    	  ReadMode      -> (#const O_RDONLY)
	    	  WriteMode     -> (#const O_WRONLY)
	    	  ReadWriteMode -> (#const O_RDWR)
	    	  AppendMode    -> (#const O_APPEND)

		execute :: SQLite3 -> String -> IO Integer
		execute sqlite query =
		  withCString query $ \pQuery -> do
		    sqlite3_exec sqlite pQuery nullFunPtr nullPtr nullPtr >>= handleSqlResult sqlite
		    changes <- sqlite3_changes sqlite
		    return $! fromIntegral changes

		prepare :: Connection -> SQLite3 -> String -> IO Statement
		prepare connection sqlite query =
		  withCString query $ \cstr ->
		  alloca $ \ppStmt -> do
		    sqlite3_prepare sqlite cstr (-1) ppStmt nullPtr >>= handleSqlResult sqlite
		    stmt <- peek ppStmt
		    count <- sqlite3_column_count stmt
		    defs <- getFieldDefs stmt 0 count
		    refFalse <- newMVar False
		    return (Statement
		    		{ stmtConn   = connection
		    		, stmtClose  = sqlite3_finalize stmt >>= handleSqlResult sqlite
		    		, stmtExecute= executePrepared defs sqlite stmt
		    		, stmtFetch  = fetch stmt
		    		, stmtGetCol = getColValue stmt
		    		, stmtFields = defs
		    		, stmtClosed = refFalse
				})

		query :: Connection -> SQLite3 -> String -> IO Statement
		query connection sqlite query =
		  withCString query $ \cstr ->
		  alloca $ \ppStmt -> do
		    sqlite3_prepare sqlite cstr (-1) ppStmt nullPtr >>= handleSqlResult sqlite
		    stmt <- peek ppStmt
		    count <- sqlite3_column_count stmt
		    defs <- getFieldDefs stmt 0 count
		    refFalse <- newMVar False
		    return (Statement
				{ stmtConn   = connection
				, stmtClose  = sqlite3_finalize stmt >>= handleSqlResult sqlite
				, stmtExecute= throwDyn SqlUnsupportedOperation
				, stmtFetch  = fetch stmt
				, stmtGetCol = getColValue stmt
				, stmtFields = defs
				, stmtClosed = refFalse
				})

		executePrepared [] sqlite stmt values = do
		  sqlite3_reset stmt >>= handleSqlResult sqlite
		  setParams sqlite stmt 1 values
		  sqlite3_step stmt
		  changes <- sqlite3_changes sqlite
		  return $! fromIntegral changes
		executePrepared _  sqlite stmt values = do
		  sqlite3_reset stmt >>= handleSqlResult sqlite
		  setParams sqlite stmt 0 values
		  return (-1)

		setParams sqlite stmt index []                                 = return ()
		setParams sqlite stmt index (SqlStringValue    value : values) = do
		  withCStringLen value $ \(cstr,len) -> do
		    sqlite3_bind_text stmt index cstr (fromIntegral len) (#const SQLITE_TRANSIENT) >>= handleSqlResult sqlite
		  setParams sqlite stmt (index+1) values
		setParams sqlite stmt index (SqlIntValue       value : values) = do
		  sqlite3_bind_int stmt index (fromIntegral value) >>= handleSqlResult sqlite
		  setParams sqlite stmt (index+1) values
		setParams sqlite stmt index (SqlDoubleValue    value : values) = do
		  sqlite3_bind_double stmt index value >>= handleSqlResult sqlite
		  setParams sqlite stmt (index+1) values
		setParams sqlite stmt index (SqlBoolValue      value : values) = do
		  alloca $ \pchar -> do
		    poke pchar (castCharToCChar $! (if value then 't' else 'f'))
		    sqlite3_bind_text stmt index pchar 1 (#const SQLITE_TRANSIENT) >>= handleSqlResult sqlite
		  setParams sqlite stmt (index+1) values
		setParams sqlite stmt index (SqlClockTimeValue value : values) = throwDyn SqlUnsupportedOperation
		setParams sqlite stmt index (SqlNullValue            : values) = do
		  sqlite3_bind_null stmt index >>= handleSqlResult sqlite
		  setParams sqlite stmt (index+1) values

		getFieldDefs :: SQLite3Stmt -> CInt -> CInt -> IO [FieldDef]
		getFieldDefs stmt index count
		  | index >= count = return []
		  | otherwise      = do name <- sqlite3_column_name stmt index >>= peekCString
                                        defs <- getFieldDefs stmt (index+1) count
		                        return ((name,SqlText,True):defs)

		tables :: Connection -> SQLite3 -> IO [String]
		tables connection sqlite = do
			stmt <- query connection sqlite "select tbl_name from sqlite_master"
			collectRows (\stmt -> getFieldValue stmt "tbl_name") stmt

		describe :: Connection -> SQLite3 -> String -> IO [FieldDef]
		describe connection sqlite table = do
			stmt <- query connection sqlite ("pragma table_info("++table++")")
			collectRows getRow stmt
			where
				getRow stmt = do
					name <- getFieldValue stmt "name"
					notnull <- getFieldValue stmt "notnull"
					return (name, SqlText, notnull=="0")

		fetch stmt = do
		  res <- sqlite3_step stmt
		  return $! (res /= (#const SQLITE_DONE))

		getColValue stmt colNumber fieldDef f = do
		  cstr <- sqlite3_column_text stmt (fromIntegral colNumber)
		  if cstr == nullPtr
		    then f fieldDef cstr 0
		    else do strLen <- strlen cstr
		            f fieldDef cstr (fromIntegral strLen)
